
import click
import typing
import logging

from adbutils import adb, AdbDevice
from tempfile import TemporaryDirectory
from pathlib import Path

from adb_install_cert.utils import prepare_pem, open_root_shell, device_is_rooted, get_android_version


REMOTE_OUTPUT_DIR = Path("/data/local/tmp")
REMOTE_TEMP_CACERTS_DIR = REMOTE_OUTPUT_DIR / "cacerts"
REMOTE_CACERTS_DIR = Path("/system/etc/security/cacerts")
APEX_CACERTS_DIR = Path("/apex/com.android.conscrypt/cacerts")


def main():
    adb_install_cert()


@click.command()
@click.option("--pem-cert", "pem_cert_filename", help="set path to PEM certificate", required=True, type=click.Path(exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.option("--device-serial", help="select device by serial", default=None)
@click.option("--clean-remote-temp/--no-clean-remote-temp", help="remove temporary files after complete installation", is_flag=True, show_default=True, default=True)
def adb_install_cert(pem_cert_filename, device_serial, clean_remote_temp):
    logging.getLogger().setLevel(logging.DEBUG)

    device = adb.device(serial=device_serial)

    if not device_is_rooted(device):
        logging.critical(f"Root device required. Recheck connected device ${device}")
        exit(1)

    with TemporaryDirectory() as tmp_dir:
        local_prepared_pem_filename = prepare_pem(pem_cert_filename, Path(tmp_dir))
        logging.info(f"[+] prepare certificate: ${local_prepared_pem_filename}")

        remote_prepared_pem_filename = REMOTE_OUTPUT_DIR / local_prepared_pem_filename.name
        device.sync.push(local_prepared_pem_filename, str(remote_prepared_pem_filename))
        logging.info(f"[+] push prepared certificate: ${local_prepared_pem_filename}")

    android_version = get_android_version(device)

    if android_version <= 9:
        legacy_permanent_install_cert(device, remote_prepared_pem_filename)
    
    # applying new installation method to old devices too
    # this allow immediately get effect
    new_temporary_install_cert(device, remote_prepared_pem_filename)

    if android_version >= 14:
        reload_runtime_cert_store(device, remote_prepared_pem_filename)

    if clean_remote_temp:
        with open_root_shell(device) as perform_as_root:
            perform_as_root(["rm", "-r", str(REMOTE_TEMP_CACERTS_DIR)])
            perform_as_root(["rm", str(remote_prepared_pem_filename)])
        logging.info("[+] clean temporary files")


def legacy_permanent_install_cert(device: AdbDevice, remote_pem: Path):
    with open_root_shell(device) as sudo:
        sudo(["mount", "-o", "rw,remount,rw", "/"])
        sudo(["mv", str(remote_pem), str(REMOTE_CACERTS_DIR)])
        sudo(["chmod", "644",  str(REMOTE_CACERTS_DIR / remote_pem.name)])
        sudo(["mount", "-o", "ro,remount,ro", "/"])
        logging.info("[+] permanent install complete")


def new_temporary_install_cert(device: AdbDevice, remote_pem: Path):
    with open_root_shell(device) as sudo:
        sudo(["mkdir", "-p", "-m", "700", str(REMOTE_TEMP_CACERTS_DIR)])
        sudo(["cp", str(REMOTE_CACERTS_DIR / "*"), str(REMOTE_TEMP_CACERTS_DIR)])
        sudo(["mount", "-t", "tmpfs", "tmpfs", str(REMOTE_CACERTS_DIR)])
        sudo(["cp", str(REMOTE_TEMP_CACERTS_DIR / "*"), str(REMOTE_CACERTS_DIR)])
        sudo(["mv", str(remote_pem), str(REMOTE_CACERTS_DIR)])
        sudo(["chmod", "644", str(REMOTE_CACERTS_DIR / remote_pem.name)])
        logging.info("[+] temporary install complete")


def reload_runtime_cert_store(device: AdbDevice, remote_pem: Path):
    with open_root_shell(device) as sudo:
        sudo(["chcon", "u:object_r:system_file:s0", str(REMOTE_CACERTS_DIR / "*")])
        zygote_pid = int(device.shell("pidof zygote || pidof zygote64"))
        sudo([
            "nsenter", f"--mount=/proc/{zygote_pid}/ns/mnt", "--",
            "/bin/mount", "--bind", str(REMOTE_CACERTS_DIR), str(APEX_CACERTS_DIR)
        ])
        zygote_childs = [int(pid) for pid in device.shell(f"ps -o 'PID' -P {zygote_pid} | grep -v PID").splitlines()]
        for zygote_child in zygote_childs:
            sudo([
                "nsenter", f"--mount=/proc/{zygote_child}/ns/mnt", "--",
                "/bin/mount", "--bind", str(REMOTE_CACERTS_DIR), str(APEX_CACERTS_DIR),
            ])


if __name__ == "__main__":
    main()


import logging
import os
import subprocess
import typing
from contextlib import contextmanager
from adbutils._utils import adb_path
from pathlib import Path
from adbutils import AdbDevice


def prepare_pem(pem_cert_filename: Path, out_dir_path: Path) -> Path:
    pem_filename = get_pem_info(pem_cert_filename, fields=["-subject_hash_old"]).strip() + ".0"
    pem_cert_contet = get_pem_info(pem_cert_filename, keep_cert = True)
    pem_additional_content = get_pem_info(pem_cert_filename, fields=["-text", "-fingerprint"])
    pem_out_path = out_dir_path / pem_filename
    with pem_out_path.open(mode="w") as pem_out:
        pem_out.writelines([pem_cert_contet, pem_additional_content])
    return pem_out_path


def get_pem_info(pem_cert_filename: Path, fields: typing.List[str] = [], keep_cert: bool = False):
    keep_cert = [] if keep_cert else ["-noout"]
    cert_proc_result = subprocess.run(
        [
            "openssl", "x509", "-inform", "PEM",
            *fields, *keep_cert,
            "-in", str(pem_cert_filename)],
        capture_output=True, text=True, check=True
    )
    return cert_proc_result.stdout


# @contextmanager
# def open_root_shell(device: AdbDevice):
#     root_sh_cmd = [adb_path(), "-s", device.serial] if device.serial else [adb_path()]
#     root_sh_cmd.extend(["shell", "su"])
#     with subprocess.Popen(
#         root_sh_cmd,
#         stdout=subprocess.PIPE,
#         stderr=subprocess.PIPE,
#         stdin=subprocess.PIPE,
#         text=True,
#         bufsize=1
#     ) as root_sh:
#         os.set_blocking(root_sh.stdout.fileno(), False)
#         os.set_blocking(root_sh.stderr.fileno(), False)
#         def perform_command_as_root(cmd: typing.List[str]):
#             cmd = list2cmdline(cmd)

#             logging.debug(f"perform_command_as_root: {cmd}")
#             root_sh.stdin.write(cmd + "\n")
#             root_sh.stdin.flush()

#             stdout = root_sh.stdout.readlines() 
#             stderr = root_sh.stderr.readlines()
#             print(stdout, stderr)

#             return stdout, stderr

#         yield perform_command_as_root

@contextmanager
def open_root_shell(device):
    def perform_command_as_root(cmd: typing.List[str]):
        root_sh_cmd = [adb_path(), "-s", device.serial] if device.serial else [adb_path()]
        root_sh_cmd.extend(["shell", "su"])
        with subprocess.Popen(
            root_sh_cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            stdin=subprocess.PIPE,
            text=True,
            bufsize=1
        ) as root_sh:
            cmd = list2cmdline(cmd)
            logging.debug(f"perform_command_as_root: {cmd}")
            stdin, stderr = root_sh.communicate(cmd)
            return stdin, stderr
    yield perform_command_as_root


def list2cmdline(args: typing.List[str]):
    return " ".join(args)


def device_is_rooted(device: AdbDevice) -> bool:
    return len(device.shell(["which", "su"]).strip()) > 0


def get_android_version(device: AdbDevice) -> int:
    return int(device.prop.get("ro.build.version.release"))